from .._moransi import morans_i_image, morans_i_adjacency_matrix
from scipy.ndimage import generate_binary_structure
from libpysal.weights import lat2W
import numpy as np

np.random.seed(0)
im_rand = np.random.random((99, 99))

im_check = (np.arange(99 * 99) % 2 - 1).reshape(99, -1)

im_half = (np.arange(99 * 99) < 99 * 99 / 2 - 1).reshape(99, -1)


def get_adj_metric(im):
    adj_mat = np.zeros((np.prod(im.shape),) * 2)
    W = lat2W(*im.shape)
    adj_mat = W.full()[0].astype(bool)
    metric = im.flatten()
    return adj_mat, metric


def test_morans_i_image_rand():
    kernel = generate_binary_structure(2, 1)
    kernel[1, 1] = False
    assert np.isclose(
        morans_i_image(im_rand, kernel=kernel), -0.007927948045819038
    )


def test_morans_i_image_check():
    kernel = generate_binary_structure(2, 1)
    kernel[1, 1] = False
    assert np.isclose(
        morans_i_image(im_check, kernel=kernel), -0.9999999999999994
    )


def test_morans_i_image_check_inverted():
    kernel = [
        [1, 0, 1],
        [0, 0, 0],
        [1, 0, 1],
    ]
    assert np.isclose(morans_i_image(im_check, kernel=kernel), 1)


def test_morans_i_image_half():
    kernel = generate_binary_structure(2, 1)
    kernel[1, 1] = False
    assert np.isclose(
        morans_i_image(im_half, kernel=kernel), 0.9897289391169648
    )


def test_morans_i_graph_rand():
    adj_mat, metric = get_adj_metric(im_rand)
    assert np.isclose(
        morans_i_adjacency_matrix(adj_mat, metric), -0.007927948045819042
    )


def test_morans_i_graph_check():
    adj_mat, metric = get_adj_metric(im_check)
    assert np.isclose(
        morans_i_adjacency_matrix(adj_mat, metric), -0.9999999999999994
    )


def test_morans_i_graph_half():
    adj_mat, metric = get_adj_metric(im_half)
    assert np.isclose(
        morans_i_adjacency_matrix(adj_mat, metric), 0.9897289391169648
    )



import numpy as np
from scipy.ndimage import convolve
from scipy.sparse import sparray


def morans_i_image(image: np.ndarray, kernel: np.ndarray) -> float:
    """Compute and return the Moran Index of `image` using `kernel` for the weights

    Parameters
    ----------
    image : np.ndarray
        The image to compute the moran index on
    kernel : np.ndarray
        The kernel from which the weights are computed from

    Returns
    -------
    float
        The moran index of the image `image` given the weights `kernel`
    """
    image_ = image - np.mean(image)
    nb_neighb = 1 / convolve(
        np.ones_like(image, dtype=float), kernel, mode="constant"
    )
    C = (nb_neighb) * convolve(image_, kernel, mode="constant")
    M = np.nansum(image_ * C) / np.nansum(image_**2)
    return M


def morans_i_adjacency_matrix(
    adj_mat: np.ndarray | sparray, metric: np.ndarray
) -> float:
    """Compute and returns the Moran Index of a given graph and a given metric on the nodes

    The graph is expected as an adjacency matrix `adj` where `adj[i, j]`
    is the weight of the link between the nodes `i` and `j`.

    The metric is expected to be an array `m` where `m[i]` is the value
    of the metric for the node `i`.

    Parameters
    ----------
    adj_mat : np.ndarray or sparray of size N x N
        The adjacency matrix of the graph that has `N` nodes
    metric : np.ndarray of size N
        The array where are stored the metrics

    Returns
    -------
    float
        The moran index of the image `image` given the weights `kernel`
    """
    avg = np.mean(metric)
    metric_ = metric - avg

    return (
        np.sum((adj_mat @ metric_) * metric_ / adj_mat.sum(axis=0))
    ) / np.sum(metric_**2)
